:- module debug.
:- interface.
:- import_module game.
:- import_module io.

	% dump the game state to stderr, for debugging
:- pred output_state(game::in, io__state::di, io__state::uo) is det.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, map, list.

output_state(Game) -->
	io__stderr_stream(Stderr),
	io__set_output_stream(Stderr, Old),
	output_board(Game),
	%output_packages(Game^packages),
	%output_enemies(Game^enemy_robots),
	%output_my_robot(Game^my_robot),
	nl,
	io__set_output_stream(Old, _).

output_board(Game) -->
	{ y(Height) = Game^board^height },
	output_rows(y(Height + 1), Game).

output_rows(y(Y), Game) -->
	( { Y >= 0 } ->
		output_row(y(Y), Game),
		output_rows(y(Y - 1), Game)
	;
		[]
	).

output_row(Y, Game) -->
	{ x(Width) = Game^board^width },
	output_cols(x(0), x(Width + 1), Y, Game).

:- pred output_cols(x, x, y, game, io__state, io__state).
output_cols(x(X), x(MaxX), Y, Game) -->
	( { X =< MaxX } ->
		output_square(x(X), Y, Game),
		output_cols(x(X + 1), x(MaxX), Y, Game)
	;
		nl
	).

	% A square can contain multiple things --
	% Multiple robots (at the start of the game different robots
	% can occupy the same square), packages, and perhaps a home base.
	% We can't display the full contents in a single character.
	% So we just pick the most important thing to say -- the first
	% in this list.  We also use uppercase/lowercase to distinguish between
	% things on a home base and things on a plain square.
:- type contents
	--->	me
	;	enemy
	;	package
	;	maybe_package
	;	none.

/*###61 [cc] Inferred :- pred output_square((game:x), (game:y), (game:game), (io:state), (io:state)).%%%*/
output_square(X, Y, Game) -->
	{ Posn = (X,Y) },
	{ MyPosn = Game^my_robot^my_posn },
	{ EnemyPosns = map((func(E) = E^enemy_posn),
		Game^enemy_robots^values) },
	{ OnGroundPackagePosns = filter_map(package_on_ground,
		Game^packages^values) },
	{ DroppedByEnemyPackagePosns = filter_map(package_dropped_by_enemy,
		Game^packages^values) },
	{ Square = square(Game^board, X, Y) },
	{ Posn = MyPosn ->
		Contents = me
	; list__member(Posn, EnemyPosns) ->
		Contents = enemy
	; list__member(Posn, OnGroundPackagePosns) ->
		Contents = package
	; list__member(Posn, DroppedByEnemyPackagePosns) ->
		Contents = maybe_package
	;
		Contents = none
	},
	io__write_char(draw(Square, Contents)).

:- func package_on_ground(package_info) = coord is semidet.
package_on_ground(PackageInfo) = Posn :-
	PackageInfo ^ package_info_posn = on_the_ground(Posn).

:- func package_dropped_by_enemy(package_info) = coord is semidet.
package_dropped_by_enemy(PackageInfo) = Posn :-
	PackageInfo ^ package_info_posn = dropped_by_enemy(Posn).

/*###92 [cc] Inferred :- func draw((game:square_kind), (debug:contents)) = character.%%%*/
draw(lethal, _) 		= ('~').
draw(wall, _)			= ('#').
draw(home, me)			= ('R').
draw(home, enemy)		= ('E').
draw(home, package)		= ('P').
draw(home, maybe_package) 	= ('M').
draw(home, none)		= ('@').
draw(plain, me)			= ('r').
draw(plain, enemy)		= ('e').
draw(plain, package)		= ('p').
draw(plain, maybe_package)	= ('m').
draw(plain, none)		= ('.').

%-----------------------------------------------------------------------------%
