:- module game.
:- interface.
:- import_module list, int, std_util, map, random.

%----------

% The complete state of the game

:- type game ---> game(
	board :: board,
	packages :: package_map,
	my_robot :: my_robot_info,
	enemy_robots :: enemy_robot_map,
	rand_seed :: random__supply
).

% Field access functions; these get/set individual packages
% or enemies within the package_map or the enemy_robot_map.
:- func enemy(robot_id, game) = enemy_robot_info.
:- func 'enemy :='(robot_id, game, enemy_robot_info) = game.

:- func package(package_id, game) = package_info.
:- func 'package :='(package_id, game, package_info) = game.

% Functions to construct and update the game state.
% These are called by main.m.

:- func init_game(board, config, server_response) = game.

:- pred update_game_state(server_response::in, game::in, game::out) is det.

:- pred update_packages(current_packages::in, game::in, game::out) is det.

%----------

% Coordinates

% Note: coordinates start at 1, not zero!
:- type coord ---> (x,y).
:- type x ---> x(int).
:- type y ---> y(int).

%----------

% The board

:- type board.

:- func make_board(x, y, list(row)) = board.
:- type row == list(square_kind).

:- func width(board) = x.
:- func height(board) = y.

:- pred valid_coord(board, coord).
:- mode valid_coord(in, in) is semidet.

:- pred valid_coord(board, x, y).
:- mode valid_coord(in, in, in) is semidet.

:- func square(board, coord) = square_kind.
:- func square(board, x, y) = square_kind.

	% These are the same as square except that if the corrdinates are
	% out of bounds then this returns wall.
	%
:- func safe_square(board, coord) = square_kind.
:- func safe_square(board, x, y) = square_kind.

:- type square_kind ---> plain ; lethal ; wall ; home.

%----------

% Packages

	% The package_map includes all known packages,
	% including those that are being carried by our robot
	% or by enemy robots.  There may be other packages
	% that we don't yet know about; these are not included in the map.
:- type package_map == map(package_id, package_info).

:- type package_id ---> package_id(int).

:- type package_info ---> package(
	package_info_id :: package_id,
	package_info_posn :: package_posn,
	% weight and destination, if known ("no" means unknown)
	package_info_weight :: maybe(weight),
	package_info_dest :: maybe(coord)		% "no" means unknown
).

:- type package_posn
	--->	carried_by_me
	;	carried_by_enemy(robot_id)
	;	on_the_ground(coord)
	;	dropped_by_enemy(coord).
			% Might be either on the ground,
			% or more likely was dropped onto destination
			% and has now disappeared.
			% XXX we should be smarter in figuring this out,
			% e.g. if we know the object's destination
			% of whether the enemy was pushed

:- type weight == int.

%----------

% Robots

:- type enemy_robot_map == map(robot_id, enemy_robot_info).

:- type robot_id ---> robot_id(int).

:- type my_robot_info ---> my_robot_info(
	my_id :: robot_id,
	my_posn :: coord,
	my_packages :: list(package_id),
	my_money :: money,
	my_capacity :: weight
	% my_history :: list(command)	% is it worth storing this?
).

:- type enemy_robot_info ---> enemy_robot_info(
	enemy_id :: robot_id,
	enemy_posn :: coord,
	enemy_packages :: list(package_id),
	enemy_history :: list(command),
	enemy_behaviour :: enemy_behaviour
).

:- type money == int.

%----------

% The server sends us a message giving the list of all the packages
% at our current location.

:- type current_packages == list(current_package).

:- type current_package --->
	current_package(
		cur_id :: package_id,
		cur_x :: x,
		cur_y :: y,
		cur_weight :: weight
	).

%----------

% Commands

:- type command ---> command(
	bid :: money,
	action :: action
).

:- type action
	--->	move(direction)
	;	pick(list(package_id))
	;	drop(list(package_id)).

	% nop_command = drop([]), i.e. it does nothing and moves nowhere.
:- func nop_command = command.

%----------

% Server responses

% The configuration consists of 
% 
%    a unique identifier, 
%    the maximum carrying capacity, 
%    the initial amount of money (used for bidding) 
%
% These are all integers and are sent on one line, separated by spaces, in the
% above order. 

:- type config --->
	config(
		config_my_id :: robot_id,
		config_my_capacity :: weight,
		config_my_money :: money
	).

:- type server_response == list(robot_update).

:- type robot_update --->
	robot_update(
		robot_update_id::robot_id,
		robot_update_actions::list(robot_action)
	).

:- type robot_action
	--->	moved(direction)
	;	picked(package_id)
	;	dropped(package_id)
	;	position(x,y)
	.

%----------

% Directions

:- type direction ---> north ; south ; east ; west.

:- func move(coord, direction) = coord.

:- func x_offset(direction) = int.
:- func y_offset(direction) = int.

%----------

% "behaviour": some kind of summary of the history,
% for use in predicting the robot's likely future moves.
% XXX not yet used
:- type enemy_behaviour ---> enemy_behaviour.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module array, exception, int, set.

%-----------------------------------------------------------------------------%
% The complete state of the game
%-----------------------------------------------------------------------------%

enemy(Id, Game) =
	Game ^ enemy_robots ^ det_elem(Id).
'enemy :='(Id, Game, EnemyInfo) =
	Game ^ enemy_robots ^ elem(Id) := EnemyInfo.

package(Id, Game) =
	Game ^ packages ^ det_elem(Id).
'package :='(Id, Game, PackageInfo) =
	Game ^ packages ^ elem(Id) := PackageInfo.

init_game(Board, Config, InitialServerResponse) = Game :-
	Config = config(MyId, MyCapacity, MyMoney),
	process_initial_server_response(MyId, InitialServerResponse,
		MyPosn, EnemyRobots),
	% initially we start with no packages,
	% and we don't know where any of the other packages are
	MyPackages = [],
	Packages = map__init,
	MyRobot = my_robot_info(MyId, MyPosn, MyPackages, MyMoney, MyCapacity),
	random__init(42, Seed),
	Game = game(Board, Packages, MyRobot, EnemyRobots, Seed).

%-----------------------------------------------------------------------------%

:- pred process_initial_server_response(robot_id, server_response, coord,
	map(robot_id, enemy_robot_info)).

process_initial_server_response(MyId, Updates, MyPosn, EnemyRobots) :-
	process_initial_server_response_2(MyId, Updates, no, MaybeMyPosn,
		map__init, EnemyRobots),
	( MaybeMyPosn = yes(MyPosn)
	; MaybeMyPosn = no, throw("no initial position")
	).

:- pred process_initial_server_response_2(robot_id, server_response,
		maybe(coord), maybe(coord), enemy_robot_map, enemy_robot_map).

process_initial_server_response_2(_MyId, [], MyPosn, MyPosn,
		EnemyRobots, EnemyRobots).
process_initial_server_response_2(MyId, [Update|Updates], MyPosn0, MyPosn,
		EnemyRobots0, EnemyRobots) :-
	Update = robot_update(Id, Actions),
	( Actions = [position(X,Y)] ->
		( Id = MyId ->
			MyPosn1 = yes((X,Y)),
			EnemyRobots1 = EnemyRobots0
		;
			EnemyId = Id,
			EnemyPosn = (X,Y),
			EnemyPackages = [],
			EnemyHistory = [],
			EnemyBehaviour = enemy_behaviour,
			EnemyRobot = enemy_robot_info(
				EnemyId, EnemyPosn, EnemyPackages,
				EnemyHistory, EnemyBehaviour),
			EnemyRobots1 = map__set(EnemyRobots0, Id, EnemyRobot),
			MyPosn1 = MyPosn0
		),
		process_initial_server_response_2(MyId, Updates, MyPosn1,
			MyPosn, EnemyRobots1, EnemyRobots)
	;
		throw("invalid initial server response")
	).

%-----------------------------------------------------------------------------%

update_game_state(Updates) -->
	list__foldl(apply_response, Updates),
	%
	% Delete dead robots.
	%
	% The spec says "A robot is listed if AND ONLY IF it was alive at
	% the beginning of the turn."  Hence any robots not listed must be
	% dead.
	%
	{ LiveRobotIds = list__map((func(U) = U^robot_update_id), Updates) },
	EnemyRobots =^ enemy_robots,
	^enemy_robots := map__select(EnemyRobots, list_to_set(LiveRobotIds)),
	Packages =^ packages,
	{ LivePackageIds = list__filter(
		package_is_live(Packages, LiveRobotIds),
		Packages ^ keys) },
	^packages := map__select(Packages, list_to_set(LivePackageIds)).

	% Fail if a package is being held by a dead robot.
	% Such packages disappear when the robot dies.
:- pred package_is_live(package_map::in, list(robot_id)::in, package_id::in)
	is semidet.
package_is_live(Packages, LiveRobots, PackageId) :-
	PackageInfo = Packages ^ det_elem(PackageId),
	\+ (
		PackageInfo ^ package_info_posn = carried_by_enemy(EnemyId),
		\+ (EnemyId `list__member` LiveRobots)
	).

:- pred apply_response(robot_update::in, game::in, game::out) is det.
apply_response(Update, GameState0, GameState) :-
	Update = robot_update(Id, Actions),
	MyId = GameState0^my_robot^my_id,
	( Id = MyId ->
		list__foldl(apply_my_action, Actions, GameState0,
			GameState)
	;
		list__foldl(apply_enemy_action(Id), Actions, GameState0,
			GameState)
	).

:- pred apply_enemy_action(robot_id::in, robot_action::in,
	game::in, game::out) is det.
apply_enemy_action(Id, position(X,Y)) -->
	^enemy(Id)^enemy_posn := (X,Y).
apply_enemy_action(Id, moved(Direction)) -->
	 OldPosn =^ enemy(Id)^enemy_posn,
	^enemy(Id)^enemy_posn := move(OldPosn, Direction).
apply_enemy_action(Id, picked(PackageId)) -->
	OldPackages =^ enemy(Id)^enemy_packages,
	^enemy(Id)^enemy_packages := [PackageId | OldPackages],
	^package(PackageId)^package_info_posn := carried_by_enemy(Id).
apply_enemy_action(Id, dropped(PackageId)) -->
	EnemyPackages =^ enemy(Id)^enemy_packages,
	EnemyPosn =^ enemy(Id)^enemy_posn,
	Packages =^ packages,
	PackageDest =^ package(PackageId)^package_info_dest,
	^enemy(Id)^enemy_packages := list__delete_all(EnemyPackages, PackageId),
	( { PackageDest = yes(EnemyPosn) } ->
		^packages := map__delete(Packages, PackageId)
	; { PackageDest = yes(_) } ->
		^package(PackageId)^package_info_posn :=
			on_the_ground(EnemyPosn)
	;
		% We don't know whether it was dropped at the destination or
		% not.  Probably it was, unless the enemy robot was
		% subject to pushing.  It's probably not worth trying for
		% these.
		^package(PackageId)^package_info_posn :=
			dropped_by_enemy(EnemyPosn)
	).

:- pred apply_my_action(robot_action::in, game::in, game::out) is det.
apply_my_action(position(X,Y)) -->
	{ NewPosn = (X,Y) },
	^my_robot^my_posn := NewPosn.
apply_my_action(moved(Direction)) -->
	OldPosn =^ my_robot^my_posn,
	^my_robot^my_posn := move(OldPosn, Direction).
apply_my_action(picked(PackageId)) -->
	OldPackages =^ my_robot^my_packages,
	^my_robot^my_packages := [PackageId | OldPackages],
	^package(PackageId)^package_info_posn := carried_by_me.
apply_my_action(dropped(PackageId)) -->
	MyPackages =^ my_robot^my_packages,
	MyPosn =^ my_robot^my_posn,
	Packages =^ packages,
	PackageDest =^ package(PackageId)^package_info_dest,
	^my_robot^my_packages := list__delete_all(MyPackages, PackageId),
	( { PackageDest = yes(MyPosn) } ->
		^packages := map__delete(Packages, PackageId)
	; { PackageDest = yes(_) } ->
		^package(PackageId)^package_info_posn := on_the_ground(MyPosn)
	;
		% PackageDest = no (i.e. destination unknown).
		% This should never happen, since we should always have
		% seen the package destination before we picked it up.
		{ throw("dropped package with unknown destination") }
	).

%-----------------------------------------------------------------------------%

update_packages(Packages, GameState0, GameState) :-
	list__foldl(update_package, Packages, GameState0, GameState).

:- pred update_package(current_package::in, game::in, game::out) is det.
update_package(current_package(Id, X, Y, Weight)) -->
	{ Destination = (X,Y) },
	CurrentPosn =^ my_robot^my_posn,
	{ Package = package(Id, on_the_ground(CurrentPosn),
		yes(Weight), yes(Destination)) },
	^package(Id) := Package.

%-----------------------------------------------------------------------------%
% The board
%-----------------------------------------------------------------------------%

:- type board == array(array(square_kind)).

width(Board) = x(Board ^ elem(0) ^ size - 2).

height(Board) = y(Board ^ size - 2).

valid_coord(Board, x(X), y(Y)) :-
	X > 0,
	Y > 0,
	x(W) = width(Board),
	X =< W,
	y(H) = height(Board),
	Y =< H.

valid_coord(Board, (X, Y)) :-
	valid_coord(Board, X, Y).

square(Board, (X,Y)) = square(Board, X, Y).

square(Board, x(X), y(Y)) = Board^elem(Y)^elem(X).

safe_square(Board, (X, Y)) = safe_square(Board, X, Y).

safe_square(Board, X, Y) = square(Board, X, Y).

make_board(x(X), _Y, Rows) = array(
	[array(list__duplicate(X + 2, wall))] ++
	list__map((func(Row) = array([wall] ++ Row ++ [wall])), Rows) ++
	[array(list__duplicate(X + 2, wall))]
).

%-----------------------------------------------------------------------------%
% Commands
%-----------------------------------------------------------------------------%

nop_command = command(1, drop([])).

%-----------------------------------------------------------------------------%
% Directions
%-----------------------------------------------------------------------------%

move((x(X),y(Y)), Direction) =
	(x(X + x_offset(Direction)), y(Y + y_offset(Direction))).

x_offset(north) = 0.
x_offset(south) = 0.
x_offset(east) = 1.
x_offset(west) = -1.

y_offset(north) = 1.
y_offset(south) = -1.
y_offset(east) = 0.
y_offset(west) = 0.

%-----------------------------------------------------------------------------%
