% The top-level

:- module main.
:- interface.
:- import_module io.

:- pred main(io__state::di, io__state::uo) is cc_multi.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module std_util, exception, list.
:- import_module game, protocol, debug, strategy.

main -->
	io__command_line_arguments(Arguments),
	( { Arguments = [Host, Port] } ->
		connect_to_server(Host, Port)
	;
		io__write_string("Usage: main <hostname> <port>\n"),
		% by default, just run from stdin/stdout
		io__write_string("Entering debug mode - I/O via stdin/stdout\n")
	),

	init(GameState0),
	run(GameState0).

init(Game) -->
	% The initialization phase occurs as soon as a client joins a game. The
	% protocol is as follows:
	%
	%  1.Client: send the string "Player"
	%  2.Server: send board
	%  3.Server: send player's configuration
	%  4.Server: send initial server response
	%    (with the initial location of all players)
	send_init,
	read_board(Board),
	read_config(Config),
	read_server_response(InitialResponse),
	{ Game = init_game(Board, Config, InitialResponse) },
	io__stderr_stream(Stderr),
	io__write_string(Stderr, "finished init\n").

% Note: this is an infinite loop.
% When the game ends, the server will close the connection,
% and we'll throw an exception.
:- mode run(in, di, uo) is cc_multi.
run(GameState0) -->
	output_state(GameState0), % for debugging
	do_turn(GameState0, GameState),
	run(GameState).

:- mode do_turn(in, out, di, uo) is cc_multi.
do_turn(GameState0, GameState) -->
	read_current_packages(Packages),
	{ update_packages(Packages, GameState0, GameState1) },
	{ choose_command(Packages, GameState1, Command, GameState2) },
	send_command(Command),
	read_server_response(Response),
	{ update_game_state(Response, GameState2, GameState) }.

:- pred choose_command(current_packages, game, command, game).
:- mode choose_command(in, in, out, out) is cc_multi.
choose_command(Packages, GameState0, Command, GameState) :-
	% first try the smart algorithm,
	% then if that fails, use the safe algorithm
	try(smart_algorithm(Packages, GameState0), ExceptionResult),
	(
		ExceptionResult = succeeded({Command, GameState})
	;
		ExceptionResult = exception(_),
		safe_algorithm(Packages, GameState0, Command),
		GameState = GameState0
	).

:- pred safe_algorithm(current_packages::in, game::in, command::out) is det.
safe_algorithm(_CurrentPackages, _GameState, Command) :-
	% XXX this is a really dumb algorithm...
	Command = nop_command.

%-----------------------------------------------------------------------------%

:- pred smart_algorithm(current_packages::in, game::in,
		{command,game}::out) is det.
smart_algorithm(CurrentPackages, GameState0, {Command, GameState}) :-
	strategy__choose_command(CurrentPackages, Command, GameState0,
		GameState).

%-----------------------------------------------------------------------------%
