%-----------------------------------------------------------------------------%
% This module implements the navigation strategy to be used by the robot.
%
% The strategy is based around an operator that takes a set of paths P
% from the current location and produces a new set of paths that can be
% reached from P by adding one extra step.
%
% For each path we collect certain information which can help us evaluate
% the path, e.g. the path length.  When two paths end at the same location,
% we prune the path which we evaluate to be the worst.
%
% If a step results in no new locations being reachable, then we have reached
% a fixpoint and we can select the best known path.  If we never remove any
% locations, then since there is a finite number of different locations on
% the board our algorithm must terminate.  However, we may wish to limit
% the number of iterations or otherwise prune the tree to improve performance.
%-----------------------------------------------------------------------------%

:- module navigate.
:- interface.
:- import_module game.
:- import_module list, set, std_util.

	% Values of this type represent paths starting from the current
	% position.  Note that the current position is implied; it is not
	% directly represented here.
	%
:- type path
	--->	path(
			path_info	:: path_info,
				% Accumulated information about the path.

			path_end	:: coord,
				% The final location.

			last_move	:: direction,
				% The final move before reaching the above
				% location.  Note carefully: we start the
				% iterations with a path that has the start
				% location as its path_end.  For this path,
				% the last_move field is meaningless.

			previous_moves	:: maybe(path)
				% If there were previous moves M then this
				% is yes(M), otherwise no.
		).

	% This type records the information that we accumulate for each
	% path.  We use this information to evaluate the path.
	%
:- type path_info
	--->	path_info(
			length		:: int,
				% The number of steps in the path, including
				% picking up and dropping packages.

			water		:: int,
				% The number of times we encounter water
				% next to us with open space directly
				% opposite.  These locations are dangerous.

			open		:: int,
				% The number of times we encounter open
				% space next to us with more open space
				% directly opposite.  At these locations,
				% we may be pushed and lose our packages.

			enemies		:: int,
				% The number of times we go near an enemy
				% robot.  This may be good or bad, depending
				% on our mood.

			packages	:: list(package_id),
				% The list of packages I would expect to be
				% carrying at the end of the path.

			delivered	:: list(package_id)
				% The list of packages I would expect to
				% have delivered by the end of the path.
		).

	% We evaluate paths in terms of what is in the path_info.  A
	% function of this type must be provided by the caller of the step
	% operator.  The reason this function is abstracted away is that
	% we may wish to use different strategies at different times.  E.g.:
	% 	- later in the game we may not consider the risk of water
	%	  to be as serious; 
	%	- when we hold valuable packages we may wish to avoid opponents
	%	  and open space.
	%	- we may wish to impose a maximum length if we don't have
	%	  much money left.
	%
:- type value_func == (func(path_info) = float).

%-----------------------------------------------------------------------------%

	% Values of this type represent sets of paths starting from the
	% current location.  This is the domain which our operator works on.
	%
:- type path_set
	--->	path_set(
			frontier	:: list(path),
				% New paths that have just been reached.

			reached		:: set(coord),
				% The set of all locations we have reached,
				% either in the current frontier or some time
				% earlier.

			best		:: maybe({path, float})
				% The best path seen so far, and its valuation.
				% If we haven't taken any steps yet, this
				% will be 'no'.
		).

%-----------------------------------------------------------------------------%

	% find_best_path(Game, ValueFunc) returns the best path and value
	% available, given the following:
	%	Game		game data
	%	ValueFunc	evaluates paths.
	% The reason we return the value along with the path is that the
	% caller may wish to adjust its overall strategy based on how good
	% the best path is (e.g. if the result is no good it may wish to
	% be more aggressive).
	%
	% If there are no paths, return 'no'.
	%
:- func find_best_path(game, value_func) = maybe({path, float}).

	% Returns the list of moves in the order they should be performed.
	%
:- func get_moves(path) = list(direction).

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, float, exception.

find_best_path(Game, ValueFunc) = Best :-
	PathSet0 = init_path_set(Game ^ my_robot),
	PathSet = iterate_step_to_fixpoint(Game, ValueFunc, PathSet0),
	Best = PathSet ^ best.

get_moves(Path) = get_moves_2(Path, []).

:- func get_moves_2(path, list(direction)) = list(direction).

get_moves_2(Path, Moves0) = Moves :-
	MaybePreviousMoves = Path ^ previous_moves,
	(
		MaybePreviousMoves = yes(PreviousMoves),
		Moves = get_moves_2(PreviousMoves, [Path ^ last_move | Moves0])
	;
		MaybePreviousMoves = no,
		% Don't insert the last_move of this path, since it is a dummy.
		Moves = Moves0
	).

%-----------------------------------------------------------------------------%

:- func init_path_set(my_robot_info) = path_set.

init_path_set(MyInfo) = PathSet :-
	Pos = MyInfo ^ my_posn,
	PathInfo = init_path_info(MyInfo),
	Dummy = north,
	Frontier = [path(PathInfo, Pos, Dummy, no)],
	PathSet = path_set(Frontier, set__init, no).

:- func init_path_info(my_robot_info) = path_info.

init_path_info(MyInfo) = path_info(0, 0, 0, 0, MyInfo ^ my_packages, []).

%-----------------------------------------------------------------------------%

:- func iterate_step_to_fixpoint(game, value_func, path_set) = path_set.

iterate_step_to_fixpoint(Game, ValueFunc, PathSet0) = PathSet :-
	Frontier0 = PathSet0 ^ frontier,
	(
		Frontier0 = [],
		%
		% Reached a fixpoint, so return.
		%
		PathSet = PathSet0
	;
		Frontier0 = [_ | _],
		%
		% Calculate the new frontier.
		%
		NewPaths0 = do_one_step(Game ^ board, Frontier0),
		list__sort_and_remove_dups(compare_first, NewPaths0, NewPaths1),
		NewPaths2 = remove_reached_locations(PathSet0 ^ reached,
			NewPaths1),
		Frontier = calculate_new_frontier(Game, NewPaths2),

		%
		% Now determine the other fields of the path set.
		%
		Reached = add_locations(PathSet0 ^ reached, Frontier),
		Best = find_best_path(ValueFunc, PathSet0 ^ best, Frontier),

		%
		% Keep iterating with the new path set.
		%
		PathSet1 = ((PathSet0 ^ best := Best)
				^ frontier := Frontier)
				^ reached := Reached,
		PathSet = iterate_step_to_fixpoint(Game, ValueFunc, PathSet1)
	).

	% From each path in the given set take one step in each direction
	% possible.  If the resulting location is not on a wall or on water,
	% add the tuple of information to the output list.
	%
	% The output list is unsorted.
	%
:- func do_one_step(board, list(path)) = list({coord, direction, path}).

do_one_step(Board, Paths) = list__foldl(do_one_step_path(Board), Paths, []).

:- func do_one_step_path(board, path, list({coord, direction, path}))
	= list({coord, direction, path}).

do_one_step_path(Board, Path, Ps0) = Ps :-
	Ps1 = do_one_step_direction(Board, north, Path, Ps0),
	Ps2 = do_one_step_direction(Board, south, Path, Ps1),
	Ps3 = do_one_step_direction(Board, east, Path, Ps2),
	Ps = do_one_step_direction(Board, west, Path, Ps3).

:- func do_one_step_direction(board, direction, path,
	list({coord, direction, path})) = list({coord, direction, path}).

do_one_step_direction(Board, Direction, Path, Ps0) = Ps :-
	Pos = move(Path ^ path_end, Direction),
	(
		square_is_free(Board, Pos)
	->
		Ps = [{Pos, Direction, Path} | Ps0]
	;
		Ps = Ps0
	).

	% square_is_free(Board, Pos) is true iff Pos is on the board and
	% the square kind there is plain or home.
:- pred square_is_free(board, coord).
:- mode square_is_free(in, in) is semidet.

square_is_free(Board, Pos) :-
	Kind = safe_square(Board, Pos),
	(
		Kind = plain
	;
		Kind = home
	).

:- pred compare_first({A, B, C}, {A, B, C}, comparison_result).
:- mode compare_first(in, in, out) is det.

compare_first({A, _, _}, {B, _, _}, Result) :-
	compare(Result, A, B).

	% remove_reached_locations(Reached, Paths) is the list of paths
	% which are members of Paths and do not have their first element
	% in Reached.  The list order must be preserved.
:- func remove_reached_locations(set(coord), list({coord, direction, path}))
	= list({coord, direction, path}).

remove_reached_locations(Reached, Ps) = list__filter(NotReached, Ps) :-
	NotReached = (pred({A, _, _}::in) is semidet :-
			\+ set__member(A, Reached)
		).

%-----------------------------------------------------------------------------%

	% Build all the path_infos.  The second argument must be sorted on
	% the first component (the coord), and must contain no duplicate
	% coords.
:- func calculate_new_frontier(game, list({coord, direction, path}))
	= list(path).

calculate_new_frontier(Game, NewPaths) = Frontier :-
	RevFrontier = calculate_new_frontier_2(Game, NewPaths, []),
	Frontier = list__reverse(RevFrontier).

:- func calculate_new_frontier_2(game, list({coord, direction, path}),
	list(path)) = list(path).

calculate_new_frontier_2(_, [], Frontier) = Frontier.
calculate_new_frontier_2(Game, NewPaths, Frontier0) = Frontier :-
	NewPaths = [{Pos, Dir, Path0} | NewPaths0],
	PathInfo = update_path_info(Game, Pos, Path0),
	Path = path(PathInfo, Pos, Dir, yes(Path0)),
	Frontier = calculate_new_frontier_2(Game, NewPaths0,
			[Path | Frontier0]).

:- func update_path_info(game, coord, path) = path_info.

update_path_info(Game, Pos, Path0) = PathInfo :-
	PathInfo0 = Path0 ^ path_info,
	PathInfo1 = PathInfo0 ^ length := PathInfo0 ^ length + 1,
	PathInfo2 = update_path_info_surrounds(Game, Pos, PathInfo1),
	PathInfo3 = update_path_info_enemies(Game, Pos, PathInfo2),
	PathInfo4 = update_path_info_home_bases(Game, Pos, PathInfo3),
	PathInfo = update_path_info_dests(Game, Pos, PathInfo4).

:- func update_path_info_surrounds(game, coord, path_info) = path_info.

update_path_info_surrounds(Game, Pos, PathInfo0) = PathInfo :-
	Board = Game ^ board,
	N = safe_square(Board, move(Pos, north)),
	S = safe_square(Board, move(Pos, south)),
	E = safe_square(Board, move(Pos, east)),
	W = safe_square(Board, move(Pos, west)),
	PathInfo1 = PathInfo0 ^ water := PathInfo0 ^ water
			+ count_water_open(N, S)
			+ count_water_open(S, N)
			+ count_water_open(E, W)
			+ count_water_open(W, E),
	PathInfo = PathInfo1 ^ open := PathInfo1 ^ open
			+ count_open_open(N, S)
			+ count_open_open(E, W).

:- func count_water_open(square_kind, square_kind) = int.

count_water_open(Water, Open) = N :-
	(
		Water = lethal,
		is_open_space(Open)
	->
		N = 1
	;
		N = 0
	).

:- func count_open_open(square_kind, square_kind) = int.

count_open_open(K1, K2) = N :-
	(
		is_open_space(K1),
		is_open_space(K2)
	->
		N = 1
	;
		N = 0
	).

:- pred is_open_space(square_kind).
:- mode is_open_space(in) is semidet.

is_open_space(plain).
is_open_space(home).

:- func update_path_info_enemies(game, coord, path_info) = path_info.

	% XXX need to do this properly.  If there are any enemies nearby,
	% add them to the total.  Maybe we should record individual ids?
update_path_info_enemies(_, _, Info) = Info.

:- func update_path_info_home_bases(game, coord, path_info) = path_info.

	% XXX need to do this properly.  If there are any known packages,
	% assume we can pick them up and add the ids to the list.
update_path_info_home_bases(_, _, Info) = Info.

:- func update_path_info_dests(game, coord, path_info) = path_info.

	% XXX need to do this properly.  If this square is a known destination
	% for a package we are carrying it, assume we drop it and add the
	% id to the delivered list.
update_path_info_dests(_, _, Info) = Info.

%-----------------------------------------------------------------------------%

	% Add the ends of the new paths to the set of reached locations.
:- func add_locations(set(coord), list(path)) = set(coord).

add_locations(Reached0, Frontier) = Reached :-
	Reached1 = set__sorted_list_to_set(list__map(get_path_end, Frontier)),
	Reached = set__union(Reached0, Reached1).

:- func get_path_end(path) = coord.

get_path_end(P) = P ^ path_end.

:- func find_best_path(value_func, maybe({path, float}), list(path))
	= maybe({path, float}).

find_best_path(_, no, []) = no.
find_best_path(ValueFunc, no, [P | Ps])
	= yes(find_best_path_2(ValueFunc, P, ValueFunc(P ^ path_info), Ps)).
find_best_path(ValueFunc, yes({Best, Val}), Ps)
	= yes(find_best_path_2(ValueFunc, Best, Val, Ps)).

:- func find_best_path_2(value_func, path, float, list(path)) = {path, float}.

find_best_path_2(_, Best, Val, []) = {Best, Val}.
find_best_path_2(ValueFunc, Best0, Val0, [P | Ps]) = BestVal :-
	ValP = ValueFunc(P ^ path_info),
	(
		ValP > Val0
	->
		BestVal = find_best_path_2(ValueFunc, P, ValP, Ps)
	;
		BestVal = find_best_path_2(ValueFunc, Best0, Val0, Ps)
	).

%-----------------------------------------------------------------------------%

