%-----------------------------------------------------------------------------%
:- module strategy.
:- interface.
:- import_module game.

%-----------------------------------------------------------------------------%

:- pred choose_command(current_packages::in, command::out,
	game::in, game::out) is det.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module list, int, random, exception, std_util.

	% XXX this is a really dumb algorithm... we should:
	%  - deliver any packages we hold for which this is the
	%    destination,
	%  - pickup packages that are here, and that we are able
	%    to carry,
	%  - decide whether we need to make a new plan, and if so then
	%    make a new plan,
	%  - look at what the current plan says we should do next,
	%  - decide on how much to bid.
	%
choose_command(CurrentPackages, Command) -->
	choose_action(CurrentPackages, Action),
	choose_bid(Action, Bid),
	{ Command = command(Bid, Action) }.

:- pred choose_action(current_packages::in, action::out,
	game::in, game::out) is det.

choose_action(CurrentPackages, Action, GameState0, GameState) :-
	(
		choose_drop(GameState0, Packages)
	->
		Action = drop(Packages),
		GameState = GameState0
	;
		choose_pick(CurrentPackages, GameState0, Packages)
	->
		Action = pick(Packages),
		GameState = GameState0
	;
		choose_move(Move, GameState0, GameState),
		Action = move(Move)
	).

%-----------------------------------------------------------------------------%

:- pred choose_drop(game::in, list(package_id)::out) is semidet.

choose_drop(Game, PackagesToDrop) :-
	CarriedIds = Game ^ my_robot ^ my_packages,
	list__filter(should_drop(Game), CarriedIds, PackagesToDrop),
	PackagesToDrop \= [].

:- pred should_drop(game::in, package_id::in) is semidet.
should_drop(Game, PackageId) :-
	MyPosn = Game ^ my_robot ^ my_posn,
	Game ^ package(PackageId) ^ package_info_dest = yes(MyPosn).

%-----------------------------------------------------------------------------%

:- pred choose_pick(current_packages::in, game::in, list(package_id)::out)
	is semidet.

choose_pick(CurrentPackages, GameState, PackagesToPickUp) :-
	% check current_packages for ones currently under us
	% that we can carry.
	list__filter_map(can_pick_up(GameState), CurrentPackages,
		PackagesToPickUp),
	PackagesToPickUp \= [].

:- pred can_pick_up(game::in, current_package::in, package_id::out) is semidet.
can_pick_up(Game, CurrentPackage, PackageId) :-
	Capacity = Game ^ my_robot ^ my_capacity,
	PackageId = CurrentPackage ^ cur_id,
	ThisWeight = package_weight(Game, PackageId), 
	CarriedIds = Game ^ my_robot ^ my_packages,
	Carrying = sum(list__map(package_weight(Game), CarriedIds)),
	Carrying + ThisWeight =< Capacity.

:- func sum(list(int)) = int.
sum(Ints) = list__foldl(func(X,Y)=X+Y, Ints, 0).

:- func package_weight(game, package_id) = weight is det.
package_weight(Game, PackageId) = Weight :-
	( Game^package(PackageId)^package_info_weight = yes(Weight0) ->
		Weight = Weight0
	;
		% This should never happen, since we should always find out
		% the package weights before we pick them up.
		throw("don't know the weight of a package that I am carrying?")
	).

%-----------------------------------------------------------------------------%

:- pred choose_move(direction::out, game::in, game::out) is det.

	% XXX decide if we need a new plan.  If so, create it. Then take
	% the first move of the current plan.
choose_move(Direction) -->
	% XXX for now, just move in a random direction
	random_direction(Direction).

:- pred random_direction(direction::out, game::in, game::out) is det.
random_direction(Direction) -->
	RS0 =^ rand_seed,
	{ random__random(R, RS0, RS) },
	^rand_seed := RS,
	{ I = R rem 4 },
	{ Direction0 = dir(I) ->
		Direction = Direction0
	;
		throw("dir failed")
	}.

:- func dir(int) = direction is semidet.
dir(0) = south.
dir(1) = north.
dir(2) = east.
dir(3) = west.

%-----------------------------------------------------------------------------%

choose_bid(_Action, Bid, GameState, GameState) :-
	% XXX should bid more if we are adjacent to other robots,
	% or bid negative if we are about to move into a square that
	% is adjacent to other robots
	Bid = 1.

%-----------------------------------------------------------------------------%
%
% Goals:
% - visit unvisited home square (to discover packages)
% - collect packages
% - deliver packages
% - push enemies
% - kill enemies
%
% - visit unvisited home square (to discover packages)
%	estimated value:
%		max(remaining capacity,
%		average of previous package values on home squares)
%	estimated cost:
%		average of previous distances from home square to dest
%		width of board
%		average distance from home square to other squares
%		
% - collect packages
%	but only if those packages are deliverable
%		(not if destination is unreachable,
%		or not reachable within our fuel limit)
%	estimated value: value of package
%	estimated cost: distance to package + 1 (to pick it up) +
%		distance to package destination + 1 (to drop it)
%
% - deliver packages
%	estimated cost: distance to distination + 1 (to drop it)
%	estimated value: value of package
%
% - push enemies
%
% - kill enemies
%
%-----------------------------------------------------------------------------%
